package share_mem

import (
	"fmt"
	"sync"
	"testing"
	"time"
)

// 共享内存并发机制

func TestCounter(t *testing.T) {
	counter := 0
	for i := 0; i < 5000; i++ {
		go func() {
			// 不安全的写法
			counter++
		}()
	}
	time.Sleep(time.Second * 1)
	t.Logf("counter= %d", counter) // 得到的结果会小于 5000
}

// Mutex的使用
func TestCounterThread(t *testing.T) {
	var mut sync.Mutex
	counter := 0
	for i := 0; i < 5000; i++ {
		go func() {
			defer func() {
				mut.Unlock()
			}()
			mut.Lock()
			counter++
		}()
	}
	time.Sleep(time.Second * 1) //去掉 time.Sleep(), counter的结果不符合预期, 小于 5000
	t.Logf("counter= %d", counter)
}

// 利用WaitGroup来确定所有任务是否结束
func TestCounterWaitGroup(t *testing.T) {
	var mut sync.Mutex
	var wg sync.WaitGroup
	counter := 0
	for i := 0; i < 5000; i++ {
		wg.Add(1)
		go func() {
			defer func() {
				mut.Unlock()
			}()
			mut.Lock()
			counter++
			wg.Done()
		}()
	}
	wg.Wait() //挂起 等待完成
	t.Logf("counter= %d", counter)
}

var rwMutex *sync.RWMutex
var wg *sync.WaitGroup

// 读写锁
func TestRWMutex(t *testing.T) {
	rwMutex = new(sync.RWMutex)
	wg = new(sync.WaitGroup)
	wg.Add(2)
	go readData(1)
	go readData(2)

	wg.Wait()
	fmt.Println("func finish")
}

func readData(i int) {
	defer wg.Done()
	fmt.Println(i, "start locking!")
	//  给读操作 上锁
	rwMutex.RLock()
	//  读数据
	fmt.Println(i, "Reading data")
	//  休眠1s
	time.Sleep(1 * time.Second)
	//  读解锁
	rwMutex.RUnlock()
	//  打印提示信息
	fmt.Println(i, "Read over")
}
